
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: CreateFrame, CreateFramePool, CreateFromMixins, Mixin, PixelUtil, UIParent

local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local MAJOR, MINOR = "SixBorder-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") end
local Library = LibStub:NewLibrary(MAJOR, MINOR)
if not Library then return end

local BORDER_SCALE = PixelUtil.GetPixelToUIUnitFactor()

local CONFIG_DEFAULTS = {
	accentBackdropInfo = {
		edgeFile = [[Interface\Buttons\WHITE8x8]],
		tile = true,
		tileSize = 8,
		tileEdge = true,
		edgeSize = 1,
	},
	accentAnchorOffset = 1,
	accentBackdropBorderColor = {1, 1, 1, 0.22},
	backgroundAnchorOffset = 0,
	backgroundBackdropColor = {0.078, 0.070, 0.082, 0.9},
	strokeBackdropInfo = {
		edgeFile = [[Interface\Buttons\WHITE8x8]],
		tile = true,
		tileSize = 8,
		tileEdge = true,
		edgeSize = 3,
	},
	strokeAnchorOffset = 0,
	strokeBackdropBorderColor = {0, 0, 0, 1},
}

-------------------------------------------------------------------------------
-- SixBorder
-------------------------------------------------------------------------------

local SixBorderMixin = {}

-- Set whether the background of the border is visible or not.
function SixBorderMixin:SetBackground(background)
	self.background:SetShown(background)
	return self
end

-- Set the background color of the border.
function SixBorderMixin:SetBackgroundColor(r, g, b, a)
	local color = r and {r, g, b, a} or self:GetConfig().backgroundBackdropColor
	self.background:SetVertexColor(unpack(color))
	self.hasCustomBackgroundColor = (r ~= nil)
	return self
end

-- Set the strata level of the background frame.
function SixBorderMixin:SetBackgroundStrata(strata)
	self.background:SetFrameStrata(strata)
	return self
end

-- Set the strata level of the border frames.
function SixBorderMixin:SetBorderStrata(strata)
	self.stroke:SetFrameStrata(strata)
	self.accent:SetFrameStrata(strata)
	return self
end

-- Set the accent color of the border.
function SixBorderMixin:SetColor(r, g, b, a)
	local color = r and {r, g, b, a} or self:GetConfig().accentBackdropBorderColor
	self.accent:SetBackdropBorderColor(unpack(color))
	self.hasCustomColor = (r ~= nil)
	return self
end

-- Set the padding of the border.
function SixBorderMixin:SetPadding(...)
	local padding = {...}
	local top, right, bottom, left

	if #padding == 0 then
		top, right, bottom, left = 0, 0, 0, 0
	elseif #padding == 1 then
		top, right, bottom, left = padding[1], padding[1], padding[1], padding[1]
	elseif #padding == 2 then
		top, right, bottom, left = padding[1], padding[2], padding[1], padding[2]
	elseif #padding == 4 then
		top, right, bottom, left = padding[1], padding[2], padding[3], padding[4]
	end

	self:ClearAllPoints()
	self:SetPoint("TOPLEFT", 0 - left, top)
	self:SetPoint("BOTTOMRIGHT", right, 0 - bottom)

	return self
end

-- Update the scale of the border based on the UI scale.
function SixBorderMixin:UpdateScale()
	self.stroke:SetScale(BORDER_SCALE)
	self.accent:SetScale(BORDER_SCALE)
	return self
end

-- Update the style of the border.
function SixBorderMixin:UpdateStyle()
	local C, B, S, A = self:GetConfig(), self.background, self.stroke, self.accent

	B:SetVertexColor(unpack(C.backgroundBackdropColor))
	S:SetBackdrop(C.strokeBackdropInfo)
	S:SetBackdropBorderColor(unpack(C.strokeBackdropBorderColor))
	A:SetBackdrop(C.accentBackdropInfo)

	if not self.hasCustomColor then
		A:SetBackdropBorderColor(unpack(C.accentBackdropBorderColor))
	end

	if B.anchorOffset ~= C.backgroundAnchorOffset then
		B.anchorOffset = C.backgroundAnchorOffset
		B:ClearAllPoints()
		B:SetPoint("TOPLEFT", self, "TOPLEFT", B.anchorOffset, -B.anchorOffset)
		B:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -B.anchorOffset, B.anchorOffset)
	end

	if S.anchorOffset ~= C.strokeAnchorOffset then
		S.anchorOffset = C.strokeAnchorOffset
		S:ClearAllPoints()
		S:SetPoint("TOPLEFT", self, "TOPLEFT", S.anchorOffset, -S.anchorOffset)
		S:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -S.anchorOffset, S.anchorOffset)
	end

	if A.anchorOffset ~= C.accentAnchorOffset then
		A.anchorOffset = C.accentAnchorOffset
		A:ClearAllPoints()
		A:SetPoint("TOPLEFT", self, "TOPLEFT", A.anchorOffset, -A.anchorOffset)
		A:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -A.anchorOffset, A.anchorOffset)
	end

	return self
end

-------------------------------------------------------------------------------
-- Factory
-------------------------------------------------------------------------------

local FactoryMixin = {}

-- Initialize a border by mixing in SixBorderMixin.
function FactoryMixin.frameInitFunc(border)
	Mixin(border, SixBorderMixin)
end

-- Reset the state of a border to its initial values.
function FactoryMixin.resetterFunc(_, border)
	border:GetParent().GetBorder = nil
	border.hasCustomBackgroundColor, border.hasCustomColor, border.parent = nil, nil, nil
	border:Hide()
	border:ClearAllPoints()
	border:SetBackground(true)
	border:SetIgnoreParentAlpha(false)
	border:SetParent(nil)
end

-- Set the configuration for the factory and update the border style of all active borders.
function FactoryMixin:SetConfig(config)
	self.config = CreateFromMixins(CONFIG_DEFAULTS, config)
	self:Update()
end

-- Create a border for the given frame.
function FactoryMixin:CreateBorder(frame, ...)
	if (not frame) then return end
	local border = frame.GetBorder and frame:GetBorder() or self.pool:Acquire()

	border.GetConfig = function() return self.config end
	border.Remove = function() self.pool:Release(border) end
	frame.GetBorder = function() return border end

	border:SetParent(frame)
	border:SetPadding(...)
	border:UpdateScale()
	border:UpdateStyle()
	border:Show()

	return border
end

-- Get the border for the given frame.
function FactoryMixin:GetBorder(frame)
	if frame and frame.GetBorder then
		return frame:GetBorder()
	end
end

-- Remove the border for the given frame.
function FactoryMixin:RemoveBorder(frame)
	if frame and frame.GetBorder then
		frame:GetBorder():Remove()
	end
end

-- Update the style of all active borders.
function FactoryMixin:Update()
	for border in self.pool:EnumerateActive() do
		border:UpdateStyle()
	end
end

-- Handle events related to display size or UI scale changes.
function FactoryMixin:OnEvent()
	BORDER_SCALE = PixelUtil.GetPixelToUIUnitFactor()

	for border in self.pool:EnumerateActive() do
		border:UpdateScale()
	end
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

-- Create a new frame factory object.
function Library:CreateFactory(config)
	return Mixin(
		CreateFrame("Frame", nil, nil, "SixBorder_FactoryTemplate"),
		FactoryMixin,
		{
			config = CreateFromMixins(CONFIG_DEFAULTS, config),
			hook = {},
			pool = CreateFramePool("FRAME", UIParent, "SixBorder_Template", FactoryMixin.resetterFunc, false, FactoryMixin.frameInitFunc),
		}
	)
end
