
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: CreateFrame, Gnosis, GnosisCharConfig, IsAddOnLoaded, math, table

local coroutine = coroutine
local Mixin = Mixin
local next = next
local UnitClass = UnitClass
local UnitIsPlayer = UnitIsPlayer
local UnitReaction = UnitReaction
local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Integration Gnosis")

M.barState = setmetatable({}, {__index = function (t, k) return {offsetCenter = 0, offsetLeft = 0, offsetRight = 0, paddingLeft = 0, paddingRight = 0} end})

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.UnitWithBar = {
	focus = true,
	mirror = true,
	player = true,
	target = true,
	vehicle = true,
}

function M:OnEnable()
	if not IsAddOnLoaded("Gnosis") then
		self:RegisterEvent("ADDON_LOADED")
		return
	end

	if Gnosis.IntroFrame then
		Gnosis.IntroFrame:Hide()
	end

	if not self:IsHooked(Gnosis, "CreateBarFrame") then
		self:RawHook(Gnosis, "CreateBarFrame", "Gnosis_CreateBarFrame")
	end

	if not self:IsHooked(Gnosis, "LoadConfig") then
		self:SecureHook(Gnosis, "LoadConfig", "UpdatePlayerBar")
	end

	if not self:IsHooked(Gnosis, "SetBarParams") then
		self:SecureHook(Gnosis, "SetBarParams", "Gnosis_SetBarParams")
	end

	if not self:IsHooked(Gnosis, "SetBorderColor") then
		self:SecureHook(Gnosis, "SetBorderColor", "Gnosis_SetBorderColor")
	end

	for bar in self:IterateBars() do
		self:StyleBar(bar)
	end

	self:RegisterEvent("LOADING_SCREEN_DISABLED", "UpdatePlayerBar")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED", "UNIT_TARGETABLE_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UNIT_TARGETABLE_CHANGED")
	self:RegisterEvent("UNIT_FACTION", "UNIT_TARGETABLE_CHANGED")
	self:RegisterEvent("UNIT_TARGETABLE_CHANGED")
	self:RegisterMessage("SIXUI_LAYOUT_CHANGED")
	self:UpdatePlayerBar()
end

function M:OnDisable()
	for bar in self:IterateBars() do
		self:UnstyleBar(bar)
	end
end

function M:IterateBars()
	return coroutine.wrap(function()
		if Gnosis then
			for _, bars in next, {Gnosis.castbars or {}, Gnosis.unusedcastbars or {}} do
				for _, bar in next, bars do
					coroutine.yield(bar)
				end
			end
		end
	end)
end

-- Set a Gnosis bar's color.
function M:SetBarColor(bar, unit, color)
	if (not bar or not bar.conf or bar.conf.unit ~= unit) then return end
	local cbconf = GnosisCharConfig.cbconf[bar.name]
	if (not cbconf) then return end
	local color = {color.r, color.g, color.b, 1}
	Mixin(cbconf.colBar, color)
	Mixin(cbconf.colBarNI, color)
	bar.bar:SetStatusBarColor(unpack(color))
end

function M:StyleBar(bar)
	if (not bar or not bar.conf) then return end
 	local iconSize = bar.icon:GetWidth() or 0
 	local iconOffsetX = math.abs(bar.conf.coord.casticon.x or 0)
 	local state = {offsetCenter = 0, offsetLeft = 0, offsetRight = 0, paddingLeft = 0, paddingRight = 0}

	if not bar._icon then
		bar._icon = CreateFrame("Frame", nil, bar)
		bar._icon:SetAllPoints(bar.icon)
	end

	if bar.conf.bIconUnlocked then
		bar._icon:Show()

		if bar.conf.iconside == "LEFT" then
			state.offsetRight = iconSize + iconOffsetX
			state.offsetCenter = (state.offsetRight / 2)
		elseif bar.conf.iconside == "RIGHT" then
			state.offsetLeft = iconSize + iconOffsetX
			state.offsetCenter = -(state.offsetLeft / 2)
		end
	else
		bar._icon:Hide()

		if bar.conf.iconside == "LEFT" then
			state.paddingLeft = iconSize
			state.offsetRight = iconSize
			state.offsetCenter = iconSize / 2
		elseif bar.conf.iconside == "RIGHT" then
			state.paddingRight = iconSize
			state.offsetLeft = iconSize
			state.offsetCenter = -(iconSize / 2)
		end
	end

	B:CreateBorder(bar, C.BORDER_PADDING, C.BORDER_PADDING + state.paddingRight, C.BORDER_PADDING, C.BORDER_PADDING + state.paddingLeft)
	B:CreateBorder(bar._icon, C.BORDER_PADDING)

	self.barState[bar.name] = state
	self:UpdateAnchors(bar.name)
end

function M:UnstyleBar(bar)
	if bar then
		B:RemoveBorder(bar)
		B:RemoveBorder(bar._icon)
	end
end

function M:UpdateAnchors(bar)
	local layout = A:GetLayout()
	local states = self.barState

	if not bar or bar == "Gnosis_Player" then
		if layout == C.Enum.Layout.Damage then
			A:RegisterFrame("SixUI_PlayerAnchor"):ClearAllPoints():SetPoint("TOP", "UIParent", "BOTTOM", states.Gnosis_Player.offsetCenter, 340 -C.BORDER_PADDING)
		elseif layout == C.Enum.Layout.Minimal then
			A:RegisterFrame("SixUI_PlayerAnchor"):ClearAllPoints():SetPoint("TOP", "UIParent", "BOTTOM", states.Gnosis_Player.offsetCenter, 430 -C.BORDER_PADDING)
		else
			A:RegisterFrame("SixUI_PlayerAnchor"):ClearAllPoints():SetPoint("TOPLEFT", "SUFUnitplayer", "TOPLEFT", -states.Gnosis_Player.offsetLeft -C.BORDER_PADDING, C.BORDER_PADDING):SetPoint("BOTTOMRIGHT", "SUFUnitplayer", "BOTTOMRIGHT", states.Gnosis_Player.offsetRight + C.BORDER_PADDING, -C.BORDER_PADDING)
		end
	end

	if not bar or bar == "Gnosis_Target" then
		A:RegisterFrame("SixUI_TargetAnchor"):ClearAllPoints():SetPoint("TOPLEFT", "SUFUnittarget", "TOPLEFT", -states.Gnosis_Target.offsetLeft -C.BORDER_PADDING, C.BORDER_PADDING):SetPoint("BOTTOMRIGHT", "SUFUnittarget", "BOTTOMRIGHT", states.Gnosis_Target.offsetRight + C.BORDER_PADDING, -C.BORDER_PADDING)
	end

	if not bar or bar == "Gnosis_Focus" then
		A:RegisterFrame("SixUI_FocusAnchor"):ClearAllPoints():SetPoint("TOPLEFT", "SUFUnitfocus", "TOPLEFT", -states.Gnosis_Focus.offsetLeft -C.BORDER_PADDING, C.BORDER_PADDING):SetPoint("BOTTOMRIGHT", "SUFUnitfocus", "BOTTOMRIGHT", states.Gnosis_Focus.offsetRight + C.BORDER_PADDING, -C.BORDER_PADDING)
	end
end

function M:UpdatePlayerBar()
	self:UNIT_TARGETABLE_CHANGED(nil, "player")
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:ADDON_LOADED(event, name)
	if name == "Gnosis" then
		self:UnregisterEvent("ADDON_LOADED")
		self:OnEnable()
	end
end

function M:SIXUI_LAYOUT_CHANGED(event, layout)
	self:UpdateAnchors()
end

-- Set a Gnosis bar's color to a unit's class or reaction color.
function M:UNIT_TARGETABLE_CHANGED(event, unit)
	local unit = unit or C.EVENT_TO_UNIT[event]
	if (not self.UnitWithBar[unit]) then return end
	local color

	if unit == "player" or unit == "vehicle" or unit == "mirror" then
		color = C.CLASS_TO_COLOR[C.PLAYER_CLASS]
	elseif UnitIsPlayer(unit) then
		color = C.CLASS_TO_COLOR[UnitClass(unit)]
	else
		color = C.REACTION_TO_COLOR[UnitReaction(unit, "player")]
	end

	for bar in self:IterateBars() do
		self:SetBarColor(bar, unit, color or C.FALLBACK_UTILITY_COLOR)
	end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- Use a hook to coerce Gnosis into reusing bars with the same name.
function M:Gnosis_CreateBarFrame(Gnosis, name, ...)
	for index, bar in next, (Gnosis.unusedcastbars or {}) do
		if bar and bar:GetName() == name then
			table.insert(Gnosis.unusedcastbars, table.remove(Gnosis.unusedcastbars, index))
		end
	end

	local bar = self.hooks[Gnosis].CreateBarFrame(Gnosis, name, ...)

	self:StyleBar(bar)

	return bar
end

-- Use a hook to restyle Gnosis bars when the icon size changes.
function M:Gnosis_SetBarParams(Gnosis, name, cfgtab, bartab)
	if Gnosis.castbars and Gnosis.castbars[name] then
		self:StyleBar(Gnosis.castbars[name])
	end
end

-- Use a hook to set SixUI's border color to a Gnosis bar's border color.
function M:Gnosis_SetBorderColor(Gnosis, bar, colBorder, colBarBg, bHideIconBorder)
	if not bar then return end

	if bar.GetBorder then
		bar:GetBorder():SetColor(unpack(colBorder))
	end

	if bar._icon and bar._icon.GetBorder then
		bar._icon:GetBorder():SetColor(unpack(colBorder))
	end
end
